/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.terremark.compute;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Locale;
import java.util.Random;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.Tag;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.MachineImage;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.compute.VirtualMachine;
import org.dasein.cloud.compute.VirtualMachineProduct;
import org.dasein.cloud.compute.VirtualMachineSupport;
import org.dasein.cloud.compute.VmState;
import org.dasein.cloud.compute.VmStatistics;
import org.dasein.cloud.dc.DataCenter;
import org.dasein.cloud.dc.Region;
import org.dasein.cloud.jclouds.terremark.Terremark;
import org.dasein.cloud.jclouds.terremark.Vdc;
import org.dasein.cloud.jclouds.terremark.network.PublicIp;
import org.jclouds.rest.AuthorizationException;
import org.jclouds.rest.RestContext;
import org.jclouds.vcloud.VCloudMediaType;
import org.jclouds.vcloud.VCloudResponseException;
import org.jclouds.vcloud.domain.ReferenceType;
import org.jclouds.vcloud.domain.Task;
import org.jclouds.vcloud.domain.TaskStatus;
import org.jclouds.vcloud.domain.VCloudExpressVApp;
import org.jclouds.vcloud.domain.VDC;
import org.jclouds.vcloud.domain.ovf.ResourceAllocation;
import org.jclouds.vcloud.domain.ovf.ResourceType;
import org.jclouds.vcloud.options.CloneVAppOptions;
import org.jclouds.vcloud.terremark.TerremarkVCloudExpressAsyncClient;
import org.jclouds.vcloud.terremark.TerremarkVCloudExpressClient;
import org.jclouds.vcloud.terremark.options.TerremarkInstantiateVAppTemplateOptions;

import com.google.common.base.Function;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;

public class VApp implements VirtualMachineSupport {
    static private final Logger logger = Logger.getLogger(VApp.class);
    
    static public URI getVAppUri(RestContext<TerremarkVCloudExpressClient, TerremarkVCloudExpressAsyncClient> ctx, String id) {
        if( id.startsWith("http") ) {
            return URI.create(id);
        }
        return URI.create(ctx.getEndpoint() + "/v" + ctx.getApiVersion() + "/vapp/" + id);   
    }
    
    private Terremark provider = null;
    
    public VApp(Terremark t) {
        provider = t;
    }
    
    @Override
    public void boot(String serverId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            try {
                ctx.getApi().powerOnVApp(getVAppUri(ctx, serverId));
            }
            catch( RuntimeException e ) {
                logger.error(e);
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public VirtualMachine clone(String serverId, String intoDcId, String name, String description, boolean powerOn, String ... firewallIds) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            try {
                CloneVAppOptions[] options = new CloneVAppOptions[powerOn ? 3 : 2];
                URI newServerId;
                Task task;
                    
                options[0] = CloneVAppOptions.Builder.withDescription(description);
                options[1] = CloneVAppOptions.Builder.deploy();
                if( powerOn ) {
                    options[2] = CloneVAppOptions.Builder.powerOn();
                }
                task = ctx.getApi().cloneVAppInVDC(Vdc.getVdcUri(ctx, intoDcId),getVAppUri(ctx, serverId), name, options);
                waitFor(task);
                newServerId = task.getOwner().getHref();
                return getVirtualMachine(newServerId.toASCIIString());
            }
            catch( RuntimeException e ) {
                logger.error(e);
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }        
    }
    
    private VirtualMachine define(String templateId, VirtualMachineProduct size, String dataCenterId, String name, boolean withAnalytics, String ... firewalls) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        
        try {
            String[] parts = size.getProductId().split(":");
            VCloudExpressVApp vapp;
            
            name = validateName(name, templateId);
            try {
                int memory = Integer.parseInt(parts[0]);
                int cpuCount = Integer.parseInt(parts[1]);
                String password = getRandomPassword();
                
                TerremarkInstantiateVAppTemplateOptions opt = TerremarkInstantiateVAppTemplateOptions.Builder.memory(memory).processorCount(cpuCount).withPassword(password);

                if( cpuCount < 1 ) {
                    cpuCount = 1;
                }
                vapp = ctx.getApi().instantiateVAppTemplateInVDC(Vdc.getVdcUri(ctx, dataCenterId), Template.getTemplateUri(ctx, templateId), name,  opt);
                if( vapp == null ) {
                    return null;
                }
                URI serverId = vapp.getHref();
                
                waitFor(ctx.getApi().deployVApp(serverId));

                VirtualMachine server = getVirtualMachine(Terremark.toId(serverId.toASCIIString()));

                server.setRootPassword(password);
                server.setProviderMachineImageId(templateId);
                return server;
            }
            catch( RuntimeException e ) {
                logger.error(e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public String getConsoleOutput(String serverId) throws InternalException, CloudException {
        return "";
    }

    @Override
    public Collection<String> listFirewalls(String serverId) throws InternalException, CloudException {
        return Collections.emptyList();
    }

    private HashMap<String,String> imageMap = new HashMap<String,String>();
    
    private String guessImageId(String osDescription) throws CloudException, InternalException {
        String searchString;

        if( osDescription == null ) {
            return null;
        }
        if( imageMap.containsKey(osDescription) ) {
            return imageMap.get(osDescription);
        }
        if( osDescription.contains("RHEL") || osDescription.contains("Red Hat") ) {
            searchString = "RHEL";
        }
        else if( osDescription.contains("CentOS") ) {
            searchString = "CentOS";
        }
        else if( osDescription.contains("Ubuntu") ) {
            searchString = "Ubuntu Server";
        }
        else if( osDescription.contains("Windows") && osDescription.contains("2003") ) {
            if( osDescription.contains("Enterprise") ) {
                searchString = "Windows 2003 Enterprise";
            }
            else {
                searchString = "Windows 2003 Standard";
            }
        }
        else if( osDescription.contains("Windows") && osDescription.contains("2008") ) {
            if( osDescription.contains("Enterprise") ) {
                searchString = "Windows Server 2008 Enterprise";
            }
            else if( osDescription.contains("Web") ) {
                searchString = "Windows Web Server 2008";
            }
            else {
                searchString = "Windows Server 2008 Standard";                
            }
        }
        else {
            searchString = osDescription;
        }
        boolean s32 = osDescription.contains("32-bit") || osDescription.contains("32 bit");
        for( MachineImage image : provider.getComputeServices().getImageSupport().listMachineImages()) {
            if( image.getDescription().startsWith(searchString) ) {
                if( s32 == image.getArchitecture().equals(Architecture.I32) ) {
                    imageMap.put(osDescription, image.getProviderMachineImageId());
                    return image.getProviderMachineImageId();
                }
            }
        }
        return null;
    }
    
    @Override
    public String getProviderTermForServer(Locale locale) {
        return "server";
    }

    static private final Random random = new Random();
    
    static public String alphabet = "ABCEFGHJKMNPRSUVWXYZabcdefghjkmnpqrstuvwxyz#@&()=+/";
    
    public String getRandomPassword() {
        StringBuilder password = new StringBuilder();
        int rnd = random.nextInt();
        int length = 8;
        
        if( rnd < 0 ) {
            rnd = -rnd;
        }
        length = length + (rnd%8);
        while( password.length() < length ) {
            char c;
            
            rnd = random.nextInt();
            if( rnd < 0 ) {
                rnd = -rnd;
            }
            c = (char)(rnd%255);
            if( alphabet.contains(String.valueOf(c)) ) {
                password.append(c);
            }
        }
        return password.toString();
    }
    
    @Override
    public VirtualMachine getVirtualMachine(String serverId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();
        
        try {
            VCloudExpressVApp server;
            
            try {
                server = ctx.getApi().getVApp(getVAppUri(ctx, serverId));
            }
            catch( RuntimeException e ) {
                String msg = e.getMessage();
                
                if( msg != null && msg.contains("NotFound") ) {
                    server = null;
                }
                else {
                    logger.error("Error loading server " + serverId + " from cloud: " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
            }
            if( server == null ) {
                return null;
            }
            return toServer(server);
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public VmStatistics getVMStatistics(String serverId, long start, long end) throws InternalException, CloudException {
        return new VmStatistics();
    }

    @Override
    public Collection<VmStatistics> getVMStatisticsForPeriod(String serverId, long start, long end) throws InternalException, CloudException {
        return Collections.emptyList();
    }

    private VirtualMachineProduct getProduct(VCloudExpressVApp server2) {
        Multimap<ResourceType,ResourceAllocation> resourceAllocation = Multimaps.index(server2.getResourceAllocations(), new Function<ResourceAllocation, ResourceType>(){

         @Override
         public ResourceType apply(ResourceAllocation arg0) {
            return arg0.getType();
         }
           
        });
        ResourceAllocation memory, processor;
        String str;
        
        if( resourceAllocation == null ) {
            return null;
        }
        Collection<ResourceAllocation> alloc = resourceAllocation.get(ResourceType.MEMORY);
        
        if( alloc != null && alloc.size() > 0 ) {
            memory = alloc.iterator().next();
        }
        else {
            memory = null;
        }
        alloc = resourceAllocation.get(ResourceType.PROCESSOR);
        if( alloc != null && alloc.size() > 0 ) {
            processor = alloc.iterator().next();
        }
        else {
            processor = null;
        }
        if( memory == null || processor == null ) {
            return null;
        }
        str = memory.getVirtualQuantity() + ":" + processor.getVirtualQuantity();
        VirtualMachineProduct product = new VirtualMachineProduct();
        
        product.setProductId(str);
        product.setName(processor.getVirtualQuantity() + " CPU, " + memory.getVirtualQuantity() + "MB RAM ");
        product.setCpuCount((int)processor.getVirtualQuantity());
        product.setRamInMb((int)memory.getVirtualQuantity());
        product.setDiskSizeInGb(1);
        product.setDescription(product.getName());
        return product;
    }
    
    static private Iterable<VirtualMachineProduct> products = null;
    
    @Override
    public Iterable<VirtualMachineProduct> listProducts(Architecture architecture) throws InternalException, CloudException {
        if( products == null ) {
            ArrayList<VirtualMachineProduct> sizes = new ArrayList<VirtualMachineProduct>();
            
            for( int ram : new int[] { 512, 1024, 1536, 2048, 4096, 8192, 12288, 16384 } ) {
                for( int cpu : new int[] { 1, 2, 4, 8 } ) {
                    VirtualMachineProduct product = new VirtualMachineProduct();

                    product.setProductId(ram + ":" + cpu);
                    product.setName(cpu + " CPU, " + ram + "M RAM");
                    product.setDescription(cpu + " CPU, " + ram + "M RAM");
                    product.setCpuCount(cpu);
                    product.setDiskSizeInGb(4);
                    product.setRamInMb(ram);
                    sizes.add(product);
                }
            }
            products = Collections.unmodifiableList(sizes);
        }
        return products;
    }

    @Override
    public VirtualMachine launch(String templateId, VirtualMachineProduct size, String dataCenterId, String serverName, String description, String withKey, String inVlanId, boolean withMonitoring, boolean forImaging, String... firewalls) throws InternalException, CloudException {
        return launch(templateId, size, dataCenterId, serverName, description, withKey, inVlanId, withMonitoring, forImaging, firewalls, new Tag[0]);
    }
    
    public VirtualMachine launch(String templateId, VirtualMachineProduct size, String dataCenterId, String serverName, String description, String withKey, String inVlanId, boolean withMonitoring, boolean forImaging, String[] firewalls, Tag ... tags) throws InternalException, CloudException {    
        VirtualMachine server = define(templateId, size, dataCenterId, serverName, withMonitoring, firewalls);
        
        boot(server.getProviderVirtualMachineId());
        return server;
    }

    @Override
    public Iterable<VirtualMachine> listVirtualMachines() throws InternalException, CloudException {
        ArrayList<VirtualMachine> servers = new ArrayList<VirtualMachine>();
        Vdc dsnVdc = (Vdc)provider.getDataCenterServices();

        for( Region region : dsnVdc.listRegions() ) {
            for( DataCenter dc : dsnVdc.listDataCenters(region.getProviderRegionId()) ) {
                Collection<ReferenceType> resources;
            
                try {
                    VDC vdc = dsnVdc.getVdc(dc.getProviderDataCenterId());
                    
                    if (vdc.getResourceEntities() != null) {
                        resources = vdc.getResourceEntities().values();
                    } else {
                        resources = Collections.emptyList();
                    }
                }
                catch( RuntimeException e ) {
                    logger.error("Could not load servers: " + e.getMessage());
                    e.printStackTrace();
                    throw new CloudException(e);
                }
                for( ReferenceType resource : resources ) { 
                    if( resource.getType().equals(VCloudMediaType.VAPP_XML) ) {
                        VirtualMachine server = getVirtualMachine(resource.getHref().toASCIIString());
                        
                        if( server != null ) {
                            servers.add(server);
                       }
                    }
                }
            }
        }
        return servers;
    }

    @Override
    public void enableAnalytics(String serverId) throws InternalException, CloudException {
        // NO-OP
    }

    @Override
    public void pause(String serverId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            try {
                ctx.getApi().powerOffVApp(getVAppUri(ctx, serverId));
            }
            catch( RuntimeException e ) {
                logger.error(e);
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void reboot(String serverId) throws CloudException, InternalException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            ctx.getApi().resetVApp(getVAppUri(ctx, serverId));
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void terminate(String serverId) throws InternalException, CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            VirtualMachine vm = getVirtualMachine(serverId);
            
            if( !vm.getCurrentState().equals(VmState.PAUSED) ) {
                waitFor(ctx.getApi().powerOffVApp(getVAppUri(ctx, serverId)));
            }
            try {
                ((PublicIp)provider.getNetworkServices().getIpAddressSupport()).clear(serverId);
                ctx.getApi().deleteVApp(getVAppUri(ctx, serverId));
            }
            catch( RuntimeException e ) {
                logger.error(e.getMessage());
                e.printStackTrace();
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void disableAnalytics(String serverId) throws InternalException, CloudException {
        // NO-OP
    }

    private VirtualMachine toServer(VCloudExpressVApp server2) throws CloudException, InternalException {
        VirtualMachine server = new VirtualMachine();
        VmState currentState;
        
        server.setProviderVirtualMachineId(Terremark.toId(server2.getHref().toASCIIString()));
        switch ( server2.getStatus() ) {
            case ON: currentState = VmState.RUNNING; break;
            case OFF: currentState = VmState.PAUSED; break;
            case SUSPENDED: currentState = VmState.PAUSED; break;
            default: currentState = VmState.PENDING; break;
        }
        server.setCurrentState(currentState);        
        server.setName(server2.getName());
        String str = server2.getOperatingSystemDescription();
        if( str == null || str.indexOf("32-bit") != -1 || str.indexOf("32 bit") != -1 ) {
            server.setArchitecture(Architecture.I32);            
        }
        else if( str.indexOf("64-bit") != -1 || str.indexOf("64 bit") != -1 ) {
            server.setArchitecture(Architecture.I64);
        }
        else {
            server.setArchitecture(Architecture.I64);
        }
        server.setCreationTimestamp(0L);
        server.setDescription(server2.getName() + " - " + server2.getOperatingSystemDescription());
        server.setPersistent(true);
        server.setPlatform(Platform.guess(server2.getOperatingSystemDescription()));
        VirtualMachineProduct size = getProduct(server2);
        
        if( size == null ) {
            return null;
        }
        else {
            server.setProduct(size);
        }
        String name = server2.getName();
        int idx = name.lastIndexOf("-");
        
        String templateId = null;
        
        if( idx != -1 && idx < (name.length()-1) ) {
            String tid = name.substring(idx+1);
            
            try {
                Integer.parseInt(tid);
                templateId = tid;
            }
            catch( NumberFormatException ignore ) {
                // ignore me
            }
        }
        if( templateId != null ) {
            server.setProviderMachineImageId(Terremark.toId(templateId));
        }
        else {
            server.setProviderMachineImageId(guessImageId(server2.getOperatingSystemDescription()));
        }
        Collection<String> addresses;
        
        try {
            addresses = server2.getNetworkToAddresses().values();
        }
        catch( RuntimeException e ) {
            logger.error(e.getMessage());
            e.printStackTrace();
            throw new CloudException(e);
        }
        if( addresses == null || addresses.size() < 1 ) {
            server.setPrivateIpAddresses(new String[0]);
        }
        else {
            String[] ips = new String[addresses.size()];
            int i = 0;
            
            for( String addr : addresses ) {
                if( i == 0 ) {
                    server.setPrivateDnsAddress(addr);
                }
                ips[i++] = addr;
            }
            server.setPrivateIpAddresses(ips);
        }
        server.setPublicDnsAddress(null);
        server.setPublicIpAddresses(new String[0]);
        server.setProviderRegionId("us-miami");
        server.setProviderDataCenterId(Terremark.toId(server2.getVDC().getHref().toASCIIString()));
        server.setProviderAssignedIpAddressId(null);
        if( server.getProviderMachineImageId() != null ) {
            server.setRootPassword(((Template)provider.getComputeServices().getImageSupport()).getPassword(server.getProviderMachineImageId()));
        }
        server.setLastPauseTimestamp(-1L);
        server.setTerminationTimestamp(-1L);
        server.setLastBootTimestamp(0L);
        server.setProviderOwnerId(provider.getContext().getAccountNumber());
        return server;
    }
    
    private String validateName(String name, String suffix) {
        int len = (suffix == null ? 0 : (1+suffix.length()));
        
        name = name.toLowerCase().replaceAll("_", "-");
        if( name.length() <= (15-len) ) {
            return (suffix == null ? name : name + "-" + suffix);
        }
        return name.substring(0, 15-(len+1)) + (suffix == null ? "" : "-" + suffix);
    }
    
    private void waitFor(Task task) throws CloudException {
        RestContext<TerremarkVCloudExpressClient,TerremarkVCloudExpressAsyncClient> ctx = provider.getCloudClient();

        try {
            TaskStatus status = task.getStatus();
            
            while( status.equals(TaskStatus.RUNNING) || status.equals(TaskStatus.QUEUED) ) {
                if( status.equals(TaskStatus.ERROR) || status.equals(TaskStatus.UNRECOGNIZED) ) {
                    throw new CloudException(task.getOwner().getName());
                }
                try { Thread.sleep(1000L); }
                catch( InterruptedException e ) { }
                task = ctx.getApi().getTask(task.getHref());
                status = task.getStatus();
            }        
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public VirtualMachineProduct getProduct(String productId) throws InternalException, CloudException {
        for( VirtualMachineProduct product : listProducts(Architecture.I64) ) {
            if( product.getProductId().equals(productId) ) {
                return product;
            }
        }
        try {
            String[] parts = productId.split(":");
            int ram = Integer.parseInt(parts[0]);
            int cpu = Integer.parseInt(parts[1]);
        
            VirtualMachineProduct product = new VirtualMachineProduct();
            
            product.setCpuCount(cpu);
            product.setRamInMb(ram);
            product.setDiskSizeInGb(4);
            product.setProductId(productId);
            product.setName(productId);
            product.setDescription(productId);
            return product;
        }
        catch( Throwable ignore ) {
            return null;
        }
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        try {
            listVirtualMachines();
            return true;
        }
        catch( AuthorizationException e ) {
            return false;
        }
        catch( VCloudResponseException e ) {
            logger.warn("Could not determine subscription status for VMs in Terremark for " + provider.getContext().getAccountNumber() + ": " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new CloudException(e);
        }
        catch( RuntimeException e ) {
            logger.warn("Could not determine subscription status for VMs in Terremark for " + provider.getContext().getAccountNumber() + ": " + e.getMessage());
            if( logger.isDebugEnabled() ) {
                e.printStackTrace();
            }
            throw new InternalException(e);            
        }
    }

    @Override
    public boolean supportsAnalytics() throws CloudException, InternalException {
        return false;
    }
}
